//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Boost Graph Library
//
// Modified by David R. Musser to use a different algorithm.
// 
// You should have received a copy of the License Agreement for the
// Boost Graph Library along with the software; see the file LICENSE.
// If not, contact Office of Research, University of Notre Dame, Notre
// Dame, IN 46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef BOOST_GRAPH_TOPOLOGICAL_SORT_HPP
#define BOOST_GRAPH_TOPOLOGICAL_SORT_HPP

#include <boost/config.hpp>
#include <boost/property_map.hpp>
#include <boost/graph/copy.hpp>
#include <queue>
#ifdef HASH
#include <hash_set>
#include <hash_map>
#else
#include <set>
#include <map>
#endif

namespace boost { 


  #ifdef HASH
  template <typename T>
  size_t identity(T p) { 
    return reinterpret_cast<size_t>(p);
  }
  #else
  template <typename T>
  const T& identity(const T& p) { 
    return p;
  }
  #endif


  // Topological Sort
  //
  // The topological sort algorithm creates a linear ordering of the
  // vertices such that if edge (u,v) appears in the graph, then u
  // comes before v in the ordering. The graph must be a directed
  // acyclic graph (DAG). Conceptually, this algorithm repeatedly finds
  // a vertex with indegree 0, removes it and all of its out edges,
  // and outputs that vertex.  To allow it to find vertices of
  // indegree 0 in constant time, it initially computes the indegree
  // of each vertex and places all vertices of indegree 0 on a queue;
  // then, when removing edges, it adds to the queue any vertex whose
  // indegree has become 0.  In actuality, the algorithm doesn't
  // remove vertices or edges from the graph; it only keeps track of
  // indegrees as though it were removing them.
  //
  // Note: this is an entirely different algorithm from the one
  // implemented in the Boost Graph Library (which is based on
  // supplying an appropriate visitor to the depth_first_search
  // algorithm).
  //
  // A hash_map is used, so that the average computing time for each
  // search is constant. (But provision is made to use a std::map
  // instead if hash tables are not included in the C++ library
  // (they're not required by the C++ standard.)

  template <typename VertexListGraph, typename OutputIterator>
  void topological_sort(VertexListGraph& g, OutputIterator result)
  {
    typedef typename graph_traits<VertexListGraph>::vertex_descriptor vertex_t;
    typedef typename graph_traits<VertexListGraph>::degree_size_type degree_size_t;
#ifdef HASH
    std::hash_map<size_t, degree_size_t> indegree;
#else
    std::map<vertex_t, degree_size_t> indegree;
#endif
    graph_traits<VertexListGraph>::vertex_iterator vi, vi_end;
    typename graph_traits<VertexListGraph>::adjacency_iterator avi, avi_end;
    for (tie(vi, vi_end) = vertices(g); vi != vi_end; ++vi)
      for (tie(avi, avi_end) = adjacent_vertices(*vi, g);
	   avi != avi_end; ++avi)
	++indegree[identity(*avi)];
    // Initialize a q with all vertices of indegree 0:
    std::queue<vertex_t> q;
    for (tie(vi, vi_end) = vertices(g); vi != vi_end; ++vi)
      if (indegree[identity(*vi)] == 0)
	q.push(*vi);
    typename graph_traits<VertexListGraph>::out_edge_iterator ei, ei_end;
    // Repeatedly remove a vertex from the queue and process it:
    while (!q.empty()) {
      vertex_t v = q.front();
      //      std::cout << "Processing vertex " << v << std::endl;
      q.pop();
      //      std::cout << " This vertex has " << out_degree(v, g) << " edges" << std::endl;
      for (tie(ei, ei_end) = out_edges(v, g);
	   ei != ei_end; ++ei) {
	// std::cout << "  processing target vertex " << target(*ei, g) << std::endl;
	if (--indegree[identity(target(*ei, g))] == 0)
	  q.push(target(*ei, g));
      }
      *result++ = v;
    }
  }

  // Subset 
  //
  // Return true if the elements in [first1, last1) are a subset of
  // those in [first2, last2); false, otherwise.  A hash_set is used,
  // so that the average computing time for each search is constant,
  // and therefore the average total time is linear. (But provision is
  // made to use a std::set instead if hash tables are not included in
  // the C++ library used to compile the code.)
  // 
  template <typename InputIterator1, typename InputIterator2>
  bool subset(InputIterator1 first1, InputIterator1 last1,
	      InputIterator2 first2, InputIterator2 last2) 
  {
    typedef typename iterator_traits<InputIterator2>::value_type value_t;
#ifdef HASH
    hash_set<size_t> set2;
#else
    set<value_t> set2;
#endif
    for (InputIterator2 k = first2; k != last2; ++k)
      set2.insert(identity(*k));
    for (InputIterator1 i = first1; i != last1; ++i)
      if (set2.find(identity(*i)) == set2.end())
	return false;
    return true;
  }

  // check_topological_sort
  // 
  // An acceptance test function for topological sort.  Returns true
  // if the elements in [result_begin, result_end) are a topological
  // ordering of the vertices of graph g, false otherwise.  In the
  // false case, this function also outputs a message to std::cout.
  //
  template <class VertexListGraph, class ForwardIterator>
  bool check_topological_sort(VertexListGraph& g, 
			      ForwardIterator result_begin, 
			      ForwardIterator result_end)
  {
    // First check that the vertex set of g and the set of elements in 
    // [result_begin, result_end) are the same. 
    typename graph_traits<VertexListGraph>::vertex_iterator vi, vi_end;
    tie(vi, vi_end) = vertices(g);
    if (!subset(result_begin, result_end, vi, vi_end)
	|| !subset(vi, vi_end, result_begin, result_end)) {
      std::cout << "***Check fails: different vertex sets" << std::endl;
      return false;
    }
    // Next check that for every edge (u, v) in g, u appears before v
    // in [result_begin, result_end)
    typedef typename graph_traits<VertexListGraph>::vertex_descriptor vertex_t;
    typedef typename graph_traits<VertexListGraph>::degree_size_type degree_size_t;
#ifdef HASH
    std::hash_map<size_t, degree_size_t> position;
#else
    std::map<vertex_t, degree_size_t> position;
#endif
    for (int i = 0; result_begin != result_end; ++i, ++result_begin)
      position[identity(*result_begin)] = i;
    typename graph_traits<VertexListGraph>::edge_iterator ei, ei_end;
    for (tie(ei, ei_end) = edges(g); ei != ei_end; ++ei)
      if (position[identity(target(*ei, g))] < position[identity(source(*ei, g))]) {
	std::cout << "***Check fails: ordering is not topological" 
		  << std::endl;
	return false;
      }
    return true;
  }

} // namespace boost

#endif /*BOOST_GRAPH_TOPOLOGICAL_SORT_H*/
